#include "bss.h"
#define MAX_CORR 300

void mean0(int n, int m, matrix a)
{
  int i,j;
  vector ave;

  ave=new_vector(n);

  for(i=0;i<n;++i){
    for(ave[i]=0,j=0;j<m;++j)ave[i]+=a[i][j];
    ave[i]/=(SCALAR)m;
  }
    
  for(i=0;i<n;++i)for(j=0;j<m;++j)a[i][j]-=ave[i];
  free_vector(ave);
}

void corr(int n, int length, matrix a, int tau, matrix cor)
{
  int i,j,k,m;
  SCALAR norm,tmp1,tmp2;

  norm = (SCALAR)(length-tau);
  m = length-tau;

  for(i=0;i<n;++i){
    for(tmp1=0,k=0;k<m;++k)tmp1+=a[i][k]*a[i][k+tau];
    cor[i][i]=tmp1/norm;
  }
  for(i=0;i<n;++i)for(j=i+1;j<n;++j){
    for(tmp1=0,tmp2=0,k=0;k<m;++k){
      tmp1+=a[i][k]*a[j][k+tau];
      tmp2+=a[j][k]*a[i][k+tau];
    }
    cor[i][j]=(tmp1+tmp2)/(2*norm);
    cor[j][i]=cor[i][j];
  }
}

void joint_diag(int n, int K, matrix *cor, matrix S)
{
  int i,j,k,ii,jj;
  vector s1,s2;
  SCALAR x,y,c,s,r,t0,t1,A,B,C;

  s1=new_vector(n);  s2=new_vector(n);
  eye(n,S);

  for(i=0;i<n-1;++i)for(j=i+1;j<n;++j){
    for(A=0,B=0,C=0,k=0;k<K;++k){
      t0=cor[k][i][i]-cor[k][j][j];
      t1=cor[k][i][j]+cor[k][j][i];
      A+=t0*t0; C+=t0*t1; B+=t1*t1;
    }
    eig2(A, B, C, &x, &y);
    r=sqrt(x*x+y*y); c=sqrt((x+r)/(2*r)); s=y/sqrt(2*r*(x+r));

    for(jj=0;jj<n;++jj){
      s1[jj]= c*S[i][jj]+s*S[j][jj];
      s2[jj]=-s*S[i][jj]+c*S[j][jj];
    }
    for(jj=0;jj<n;++jj){
      S[i][jj]=s1[jj]; S[j][jj]=s2[jj];
    }
    for(k=0;k<K;++k){
      for(jj=0;jj<n;++jj){
	s1[jj]= c*cor[k][i][jj]+s*cor[k][j][jj];
	s2[jj]=-s*cor[k][i][jj]+c*cor[k][j][jj];
      }
      for(jj=0;jj<n;++jj){
	cor[k][i][jj]=s1[jj]; cor[k][j][jj]=s2[jj];
      }
      for(ii=0;ii<n;++ii){
	s1[ii]= c*cor[k][ii][i]+s*cor[k][ii][j];
	s2[ii]=-s*cor[k][ii][i]+c*cor[k][ii][j];
      }
      for(ii=0;ii<n;++ii){
	cor[k][ii][i]=s1[ii]; cor[k][ii][j]=s2[ii];
      }
    }
  }
  free_vector(s1); free_vector(s2);
}

void sphering(int n, int length, matrix x, matrix sph)
{
  int i,j;
  SCALAR a;
  vector d,e;

  d=new_vector(n); e=new_vector(n);
  corr(n,length,x,0,sph);

  /*  tmp=new_matrix(n,n); copy_mat(n,n,sph,tmp);*/

  eigen(n,sph,d,e);

  /*  if(d[n-1]>0){*/

  for(i=0;i<n;++i){
    a=sqrt(d[i]);
    for(j=0;j<n;++j) sph[i][j]/=a;
  }
  /*}
  else{
    printf("Matirx must be singlar\n");
    for(lambda=0,i=n-1;i>=0;--i){
      if(d[i]>0){lambda=EP-d[n-1];break;}
    }
    I=new_matrix(n,n);
    zeros(n,n,I);
    for(i=0;i<n;++i) I[i][i]=lambda;
    sum_mat(n,n,tmp,I,sph); 
    eigen(n,sph,d,e);

    for(i=0;i<n;++i){
      a=sqrt(d[i]);
      for(j=0;j<n;++j) sph[i][j]/=a;
    }
    free_matrix(I);
    free_matrix(tmp);
    }*/
    
  free_vector(d); free_vector(e);
}

void bsep(int mode, int K, int n, int length, matrix x, matrix w_inv)
{
  int i,flag=0,N=n,j,ii,jj;
  matrix cor[MAX_CORR],sph,tmp,D,tmpx;
  vector pow;

  pow =new_vector(n);

  if(mode==0) mean0(n,length,x);

  for(i=0;i<K;++i) cor[i]=new_matrix(n,n);
  for(i=0;i<n;++i)for(pow[i]=0,j=0;j<length;++j) pow[i]+=x[i][j]*x[i][j];

  for(i=0;i<n;++i){
    if(flag!=0)
      for(j=0;j<length;++j)
	x[i-flag][j]=x[i][j];
    if(pow[i]==0){
      ++flag;--N;
      printf("channel %3d does not contain any data\n",i);
    }
  }
  sph =new_matrix(N,N);  D=new_matrix(N,N);
  tmp =new_matrix(N,length);

  printf("\nSphering\n");
  sphering(N, length, x, sph);
  printf("Done.\n\n");

  multiply(N,N,length,sph,x,tmp);
 
  printf("Obtaining Correlation Matrices\n");
  for(i=0;i<K;++i) corr(N,length,tmp,i+1,cor[i]);
  printf("Done.\n\n");

  printf("Rotating\n");
  joint_diag(N,K,cor,D);
  printf("Done.\n\n");

  if(flag!=0){
    tmpx=new_matrix(N,length);
    multiply(N, N, length, D, tmp, tmpx);

    for(ii=0,i=0;i<n;++i){
      if(pow[i]==0){
	for(j=0;j<length;++j) x[i][j]=0;
      }
      else{
	for(j=0;j<length;++j) x[i][j]=tmpx[ii][j];
	++ii;
      }
    }

    free_matrix(tmpx);
    mult2(N,D,sph);
    matinv(N,D,sph);
    zeros(n,n,w_inv);

    for(ii=0,i=0;i<n;++i){
      if(pow[i]!=0){
	for(jj=0,j=0;j<n;++j){
	  if(pow[j]!=0){
	    w_inv[i][j]=sph[ii][jj];
	    ++jj;
	  }
	  ++ii; 
	}
      }
    }

  }
  else{
    multiply(n, n, length, D, tmp, x);
    mult2(n,D,sph);
    matinv(n,D,w_inv);
  }
  
  free_matrix(sph); free_matrix(tmp); free_matrix(D);
  for(i=0;i<K;++i)free_matrix(cor[i]);
  free_vector(pow);
}

void separate(int src, int n, int length, matrix w_inv, matrix x, matrix out)
{
  int i;
  matrix tmp;

  tmp=new_matrix(n,length); zeros(n,length,tmp);

  for(i=0;i<length;++i)
    tmp[src][i]=x[src][i];
  multiply(n,n,length,w_inv,tmp,out);

  free_matrix(tmp);
}

SCALAR update_W(int dim,int nfft, SCALAR eta,
		matrix xf_r, matrix xf_i,
		array3 W_r, array3 W_i,
		array3 Winv_r, array3 Winv_i,SCALAR r)
{
  int i,j,f;
  static int last_dim=0,last_nfft=0;
  SCALAR det=0;
  static vector 
    z_r=NULL,z_i=NULL;
  static matrix 
    m1_r=NULL,m1_i=NULL,
    m2_r=NULL,m2_i=NULL,
    m3_r=NULL,m3_i=NULL;

  if(dim!=last_dim){
    if(m1_r!=NULL)free_matrix(m1_r); if(m1_i!=NULL)free_matrix(m1_i);
    if(m2_r!=NULL)free_matrix(m2_r); if(m2_i!=NULL)free_matrix(m2_i);
    if(m3_r!=NULL)free_matrix(m3_r); if(m3_i!=NULL)free_matrix(m3_i);
    if(z_r!=NULL) free_vector(z_r);  if(z_i!=NULL) free_vector(z_i);
    m1_r=new_matrix(dim,dim); m1_i=new_matrix(dim,dim);
    m2_r=new_matrix(dim,dim); m2_i=new_matrix(dim,dim);
    m3_r=new_matrix(dim,dim); m3_i=new_matrix(dim,dim);
    zeros(dim, dim, m1_r); zeros(dim, dim, m1_i);
    last_dim=dim;
  }
  if(nfft!=last_nfft){
    if(z_r!=NULL) free_vector(z_r);  
    if(z_i!=NULL) free_vector(z_i);
    z_r=new_vector(nfft/2+1); 
    z_i=new_vector(nfft/2+1);
    last_nfft=nfft;
  }

  for(f=0;f<nfft/2+1;++f){
    for(i=0;i<dim;++i)for(z_r[i]=0,z_i[i]=0,j=0;j<dim;++j){
      z_r[i]+=W_r[f][i][j]*xf_r[j][f]-W_i[f][i][j]*xf_i[j][f];
      z_i[i]+=W_i[f][i][j]*xf_r[j][f]+W_r[f][i][j]*xf_i[j][f];
    }
    for(i=0;i<dim;++i)for(j=0;j<dim;++j){
      if(i!=j){
	m1_r[i][j]=tanh(r*z_r[i])*r*z_r[j]+tanh(r*z_i[i])*r*z_i[j];
	m1_i[i][j]=tanh(r*z_i[i])*r*z_r[j]-tanh(r*z_r[i])*r*z_i[j];
      }
      m2_r[i][j]=Winv_r[f][i][j]; m2_i[i][j]=Winv_i[f][i][j];
      m3_r[i][j]=Winv_r[f][i][j]; m3_i[i][j]=Winv_i[f][i][j];
    }
    
    mult2(dim,m2_r,m1_r); mult2(dim,m2_i,m1_i);
    mult2(dim,m3_i,m1_r); mult2(dim,m3_r,m1_i);
    
    for(i=0;i<dim;++i){
      for(j=0;j<dim;++j){
	if(i!=j){
	  Winv_r[f][i][j]=Winv_r[f][i][j]+eta*(m2_r[i][j]-m2_i[i][j]);
	  Winv_i[f][i][j]=Winv_i[f][i][j]+eta*(m3_r[i][j]+m3_i[i][j]);
	}
	if((det=matinv_complex(dim,Winv_r[f],Winv_i[f],W_r[f],W_i[f]))==0){
	  printf("Singlar, and no update.\n");
	  break;
	}
      }
    }
  }
  return(det);
}

void W2fil_t(int dim,int nfft,int tau,
	     vector window, 
	     array3 W_r, array3 W_i,
	     array3 Winv_r, array3 Winv_i,
	     array4 filter_t)
{
  int i,j,k,f;
  static int last_tau=-1,last_nfft=0;
  SCALAR tmp_r,tmp_i;
  static vector 
    tau_fr=NULL, tau_fi=NULL,
    fil_fr=NULL, fil_fi=NULL;

  if(nfft!=last_nfft){
    if(tau_fr!=NULL) free_vector(tau_fr);
    if(tau_fi!=NULL) free_vector(tau_fi);
    if(fil_fr!=NULL) free_vector(fil_fr);
    if(fil_fi!=NULL) free_vector(fil_fi);
    tau_fr=new_vector(nfft); tau_fi=new_vector(nfft);
    fil_fr=new_vector(nfft); fil_fi=new_vector(nfft);
    last_nfft=nfft;
  }

  if(tau!=last_tau){
    for(f=0;f<nfft;++f){
      tau_fr[f]=0;
      tau_fi[f]=0;
    }
    tau_fr[tau]=1; 
    fft(-nfft,tau_fr,tau_fi);
    last_tau=tau;
  }

  for(k=0;k<dim;++k){
    for(i=0;i<dim;++i){
      for(j=0;j<dim;++j){
	for(f=0;f<nfft/2+1;++f){
	  tmp_r=Winv_r[f][i][k]*W_r[f][k][j]-Winv_i[f][i][k]*W_i[f][k][j];
	  tmp_i=Winv_i[f][i][k]*W_r[f][k][j]+Winv_r[f][i][k]*W_i[f][k][j];
	  fil_fr[f]=tmp_r*tau_fr[f]-tmp_i*tau_fi[f];
	  fil_fi[f]=tmp_i*tau_fr[f]+tmp_r*tau_fi[f];
	}
	for(f=nfft/2+1;f<nfft;++f){
	  fil_fr[f]= fil_fr[nfft-f];
	  fil_fi[f]=-fil_fi[nfft-f];
	}
	for(fft(nfft,fil_fr,fil_fi),f=0;f<nfft;++f)
	  filter_t[k][i][j][f]=window[f]*fil_fr[nfft-1-f];
      }
    }
  }
}

void W2fil_f(int dim,int nfft,int tau,
	     array3 W_r, array3 W_i,
	     array3 Winv_r, array3 Winv_i,
	     array4 filter_fr, array4 filter_fi)
{
  int i,j,k,f;

  for(k=0;k<dim;++k){
    for(i=0;i<dim;++i){
      for(j=0;j<dim;++j){
	for(f=0;f<nfft/2+1;++f){
	  filter_fr[k][i][j][f]
	    =Winv_r[f][i][k]*W_r[f][k][j]-Winv_i[f][i][k]*W_i[f][k][j];
	  filter_fi[k][i][j][f]
	    =Winv_i[f][i][k]*W_r[f][k][j]+Winv_r[f][i][k]*W_i[f][k][j];
	}
      }
    }
  }
}

int corr_f(int dim, int nfft, int shift, int len, int mnum,
	   vector window, matrix x, array4 Corr_r, array4 Corr_i)
{
  int i,j,l,f,t,tt,slen;
  vector tmphead_r,tmphead_i,tmptail_r,tmptail_i;
  matrix ptr_mat_r, ptr_mat_i,Aver_r,Aver_i;
  array3 buff_r,buff_i,head_r,head_i,tail_r,tail_i;

  slen = spl(len, nfft, shift);  

  tmphead_r = new_vector(dim);       tmphead_i = new_vector(dim);
  tmptail_r = new_vector(dim);       tmptail_i = new_vector(dim);
  Aver_r = new_matrix(nfft/2+1,dim); Aver_i = new_matrix(nfft/2+1,dim);
  zeros(nfft/2+1,dim,Aver_r);        zeros(nfft/2+1,dim,Aver_i);
  buff_r = new_array3(mnum,dim,nfft);buff_i = new_array3(mnum,dim,nfft);
  head_r = new_array3(mnum,dim,nfft/2+1);
  head_i = new_array3(mnum,dim,nfft/2+1);
  tail_r = new_array3(mnum,dim,nfft/2+1);
  tail_i = new_array3(mnum,dim,nfft/2+1);

  for(t=0;t<mnum;++t)for(f=0;f<nfft/2+1;++f){
    zeros(dim, dim, Corr_r[t][f]);
    zeros(dim, dim, Corr_i[t][f]);
  }

  printf("Making Correlation Matrixes.\n");

  for(t=0;t<mnum;++t){
    for(i=0;i<dim;++i){
      for(l=0;l<nfft;++l){
	buff_r[t][i][l] = window[l]*x[i][t*shift+l];
	buff_i[t][i][l] = 0;
      }
      fft(nfft,buff_r[t][i],buff_i[t][i]);
      for(f=0;f<nfft/2+1;++f){
	head_r[t][i][f] = buff_r[t][i][f];
	head_i[t][i][f] = buff_i[t][i][f];
	Aver_r[f][i]   += buff_r[t][i][f];
	Aver_i[f][i]   += buff_i[t][i][f];
      }
    }
  }

  for(t=0;t<mnum;++t){
    for(tt=0;tt<mnum-t;++tt){
      for(f=0;f<nfft/2+1;++f){
	for(i=0;i<dim;++i){
	  for(j=0;j<dim;++j){
	    Corr_r[t][f][i][j]+=
	      buff_r[tt][i][f]*buff_r[tt+t][j][f]
	      +buff_i[tt][i][f]*buff_i[tt+t][j][f];
	    Corr_i[t][f][i][j]+=
	      buff_i[tt][i][f]*buff_r[tt+t][j][f]
	      -buff_r[tt][i][f]*buff_i[tt+t][j][f];
	  }
	}
      }
    }
  }

  for(t=mnum;t<slen;++t){
    ptr_mat_r=buff_r[0]; ptr_mat_i=buff_i[0];
    for(tt=0;tt<mnum-1;++tt){
      buff_r[tt]=buff_r[tt+1];
      buff_i[tt]=buff_i[tt+1];
    }
    buff_r[mnum-1]=ptr_mat_r; buff_i[mnum-1]=ptr_mat_i;
    for(i=0;i<dim;++i){
      for(l=0;l<nfft;++l){
	buff_r[mnum-1][i][l]=window[l]*x[i][t*shift+l];
	buff_i[mnum-1][i][l]=0;
      }
      fft(nfft,buff_r[mnum-1][i],buff_i[mnum-1][i]);
      for(f=0;f<nfft/2+1;++f){
	Aver_r[f][i]+= buff_r[mnum-1][i][f];
	Aver_i[f][i]+= buff_i[mnum-1][i][f];
      }
    }

    for(tt=0;tt<mnum;++tt){
      for(f=0;f<nfft/2+1;++f){
	for(i=0;i<dim;++i){
	  for(j=0;j<dim;++j){
	    Corr_r[tt][f][i][j]+=
	      buff_r[mnum-1-tt][i][f]*buff_r[mnum-1][j][f]
	      +buff_i[mnum-1-tt][i][f]*buff_i[mnum-1][j][f];
	    Corr_i[tt][f][i][j]+=
	      buff_i[mnum-1-tt][i][f]*buff_r[mnum-1][j][f]
	      -buff_r[mnum-1-tt][i][f]*buff_i[mnum-1][j][f];
	  }
	}
      }
    }
  }

  for(t=0;t<mnum;++t)for(i=0;i<dim;++i)for(f=0;f<nfft/2+1;++f){
    tail_r[t][i][f] = buff_r[t][i][f];
    tail_i[t][i][f] = buff_i[t][i][f];
  }
  printf("Done.\n");
  printf("Adjusting the Average.\n");

  for(f=0;f<nfft/2+1;++f){
    for(i=0;i<dim;++i){
      tmphead_r[i]=Aver_r[f][i];
      tmphead_i[i]=Aver_i[f][i];
      tmptail_r[i]=Aver_r[f][i];
      tmptail_i[i]=Aver_i[f][i];
      for(j=0;j<dim;++j){
	Corr_r[0][f][i][j]-=
	  (Aver_r[f][i]*Aver_r[f][j]+Aver_i[f][i]*Aver_i[f][j])/slen;
	Corr_i[0][f][i][j]-=
	  (Aver_i[f][i]*Aver_r[f][j]-Aver_r[f][i]*Aver_i[f][j])/slen;
      }
    }
    for(i=0;i<dim;++i){
      Corr_r[0][f][i][i]/=slen; Corr_i[0][f][i][i]=0;
      for(j=i+1;j<dim;++j){
	Corr_r[0][f][i][j] =
	  (Corr_r[0][f][i][j]+Corr_r[0][f][j][i])/(2*slen);
	Corr_i[0][f][i][j] =
	  (Corr_i[0][f][i][j]-Corr_i[0][f][j][i])/(2*slen);
      }
      for(j=0;j<i;++j){
	Corr_r[0][f][i][j] = Corr_r[0][f][j][i];
	Corr_i[0][f][i][j] =-Corr_i[0][f][j][i];
      }
    }

    for(t=1;t<mnum;++t){
      for(i=0;i<dim;++i){
	tmphead_r[i] -= head_r[t-1][i][f];
	tmphead_i[i] -= head_i[t-1][i][f];
	tmptail_r[i] -= tail_r[mnum-t][i][f];
	tmptail_i[i] -= tail_i[mnum-t][i][f];
	for(j=0;j<dim;++j){
	  Corr_r[t][f][i][j]+=
	    -(tmphead_r[i]*Aver_r[f][j]+tmphead_i[i]*Aver_i[f][i])/slen
	    -(tmptail_r[i]*Aver_r[f][j]+tmptail_i[i]*Aver_i[f][i])/slen
	    +(Aver_r[f][i]*Aver_r[f][j]+Aver_i[f][i]*Aver_i[f][i])
	    *(slen-t)/(slen*slen);
	  Corr_i[t][f][i][j]+=
	    -(tmphead_i[i]*Aver_r[f][j]-tmphead_r[i]*Aver_i[f][i])/slen
	    -(tmptail_i[i]*Aver_r[f][j]-tmptail_r[i]*Aver_i[f][i])/slen
	    +(Aver_i[f][i]*Aver_r[f][j]-Aver_r[f][i]*Aver_i[f][i])
	    *(slen-t)/(slen*slen);
	}
      }
      for(i=0;i<dim;++i){
	Corr_r[t][f][i][i]/=(slen-t); Corr_i[t][f][i][i]=0;
	for(j=i+1;j<dim;++j){
	  Corr_r[t][f][i][j]=
	    (Corr_r[t][f][i][j]+Corr_r[t][f][j][i])/(2*(slen-t));
	  Corr_i[t][f][i][j]=
	    (Corr_i[t][f][i][j]-Corr_i[t][f][j][i])/(2*(slen-t));
	}
	for(j=0;j<i;++j){
	  Corr_r[t][f][i][j] = Corr_r[t][f][j][i];
	  Corr_i[t][f][i][j] =-Corr_i[t][f][j][i];
	}
      }
    }
  }

  free_vector(tmphead_r);   free_vector(tmphead_i);
  free_vector(tmptail_r);   free_vector(tmptail_i);
  free_matrix(Aver_r);      free_matrix(Aver_i);
  free_array3(buff_r,mnum); free_array3(buff_i,mnum);
  free_array3(head_r,mnum); free_array3(head_i,mnum);
  free_array3(tail_r,mnum); free_array3(tail_i,mnum);

  printf("Done.\n");
  return(slen);
}

void sphering_c(int dim, int nfft, int mnum,
		array3 sph_r,  array3 sph_i, 
		array4 Corr_r, array4 Corr_i)
{
  int i,j,f,t;
  SCALAR tmp;
  vector eig_vec,work_sp;
  matrix eig_mat,tmp_r,tmp_i,tmp2_r,tmp2_i,spt_r,spt_i;

  printf("Sphering.\n");

  tmp_r  = new_matrix(dim,dim);
  tmp_i  = new_matrix(dim,dim);
  tmp2_r = new_matrix(dim,dim);
  tmp2_i = new_matrix(dim,dim);
  spt_r  = new_matrix(dim,dim);
  spt_i  = new_matrix(dim,dim);

  eig_mat=new_matrix(2*dim,2*dim);
  eig_vec=new_vector(2*dim);
  work_sp=new_vector(2*dim);

  for(f=0;f<nfft/2+1;++f){
    for(i=0;i<dim;++i){
      for(j=0;j<dim;++j){
	eig_mat[i][j]         = Corr_r[0][f][i][j];
	eig_mat[i+dim][j+dim] = Corr_r[0][f][i][j];
	eig_mat[i][j+dim]     =-Corr_i[0][f][i][j];
	eig_mat[i+dim][j]     = Corr_i[0][f][i][j];
      }
    }
    eigen(2*dim,eig_mat,eig_vec,work_sp);
    for(i=0;i<dim;++i){
      tmp=(SCALAR)sqrt((double)eig_vec[2*i]);
      for(j=0;j<dim;++j){
	sph_r[f][i][j] = eig_mat[2*i][j]/tmp;
	sph_i[f][i][j] =-eig_mat[2*i][j+dim]/tmp;
	spt_r[j][i] = eig_mat[2*i][j]/tmp;
	spt_i[j][i] = eig_mat[2*i][j+dim]/tmp;
      }
    }
    for(t=0;t<mnum;++t){
      mult_cmp(dim,dim,dim,sph_r[f],sph_i[f],Corr_r[t][f],Corr_i[t][f],
	       tmp2_r,tmp2_i,tmp_r,tmp_i);
      mult_cmp(dim,dim,dim,tmp2_r,tmp2_i,spt_r,spt_i,
	       Corr_r[t][f],Corr_i[t][f],tmp_r,tmp_i);
    }
  }
  printf("Done.\n");
}

void joint_diag_f(int dim, int mnum, int nfft,
		  array4 Corr_r, array4 Corr_i,
		  array3 rot_r, array3 rot_i)
{
  int i,j,k,ii,jj,f;
  vector h_r,eig_vec,work_sp;
  matrix tmp_r,tmp_i,tmp2_r,tmp2_i,tmp3,r_r,r_i,rt_r,rt_i;
  SCALAR x,y,z,c,s_r,s_i,r;


  printf("Rotation.\n");
  eig_vec = new_vector(3);
  work_sp = new_vector(3);
  h_r     = new_vector(3);
  r_r     = new_matrix(dim,dim);
  r_i     = new_matrix(dim,dim);
  rt_r    = new_matrix(dim,dim);
  rt_i    = new_matrix(dim,dim);
  tmp_r   = new_matrix(dim,dim);
  tmp_i   = new_matrix(dim,dim);
  tmp2_r  = new_matrix(dim,dim);
  tmp2_i  = new_matrix(dim,dim);
  tmp3    = new_matrix(3,3);

  for(f=0;f<nfft/2+1;++f){
    eye(dim,rot_r[f]);
    zeros(dim,dim,rot_i[f]);

    for(i=0;i<dim-1;++i){
      for(j=i+1;j<dim;++j){
	zeros(3,3,tmp3);
	for(k=0;k<mnum;++k){
	  h_r[0]=Corr_r[k][f][i][i]-Corr_r[k][f][j][j];
	  h_r[1]=Corr_r[k][f][i][j]+Corr_r[k][f][j][i];
	  h_r[2]=Corr_i[k][f][i][j]-Corr_i[k][f][j][i];
	  for(ii=0;ii<3;++ii)for(jj=0;jj<3;++jj){
	    tmp3[ii][jj]+=h_r[ii]*h_r[jj];
	  }
	}
	
	eigen(3,tmp3,eig_vec,work_sp);
	x=tmp3[0][0]; y=tmp3[0][1]; z=tmp3[0][2];

	r  = (SCALAR)sqrt((double)(x*x+y*y+z*z));
	c  = (SCALAR)sqrt((double)((x+r)/(2*r)));
	s_r= (SCALAR)y/((SCALAR)sqrt((double)(2*r*(x+r))));
	s_i= (SCALAR)-z/((SCALAR)sqrt((double)(2*r*(x+r))));
	
	eye(dim,r_r); zeros(dim,dim,r_i);
	r_r[i][i]= c;
	r_r[i][j]= s_r; r_i[i][j]=-s_i;
	r_r[j][i]=-s_r; r_i[j][i]=-s_i;
	r_r[j][j]= c;
	for(ii=0;ii<dim;++ii)for(jj=0;jj<dim;++jj){
	  rt_r[ii][jj] = r_r[jj][ii];
	  rt_i[ii][jj] =-r_i[jj][ii];
	}

	mult_cmp(dim,dim,dim,r_r,r_i,rot_r[f],rot_i[f],
		 tmp_r,tmp_i,tmp2_r,tmp2_i);
	copy_mat(dim,dim,tmp_r,rot_r[f]);
	copy_mat(dim,dim,tmp_i,rot_i[f]);

	for(k=0;k<mnum;++k){
	  mult_cmp(dim,dim,dim,r_r,r_i,
		   Corr_r[k][f],Corr_i[k][f],tmp_r,tmp_i,tmp2_r,tmp2_i);
	  mult_cmp(dim,dim,dim,tmp_r,tmp_i,
		   rt_r,rt_i,Corr_r[k][f],Corr_i[k][f],tmp2_r,tmp2_i);
	}
      }
    }
  }

  free_vector(eig_vec); free_vector(work_sp); free_vector(h_r);

  free_matrix(r_r);    free_matrix(r_i);
  free_matrix(rt_r);   free_matrix(rt_i);
  free_matrix(tmp_r);  free_matrix(tmp_i);
  free_matrix(tmp2_r); free_matrix(tmp2_i);
  free_matrix(tmp3);
  printf("Done.\n");
}

void sph_rot2Wsep(int dim, int nfft, 
		  array3 sph_r, array3 sph_i, 
		  array3 rot_r, array3 rot_i, 
		  array3 Wsep_r, array3 Wsep_i)
{
  int f;
  matrix tmp_i,tmp_r;
  
  tmp_r = new_matrix(dim,dim);
  tmp_i = new_matrix(dim,dim);

  for(f=0;f<nfft/2+1;++f)
    mult_cmp(dim,dim,dim,rot_r[f],rot_i[f],sph_r[f],sph_i[f],
	     Wsep_r[f], Wsep_i[f],tmp_r,tmp_i);

  free_matrix(tmp_r);  free_matrix(tmp_i);
}

