#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bss.h"
#include "spec.h"

#ifndef ARRAY
#define ARRAY 1
#include "array.h"
#endif

#ifndef WAV
#define WAV 1
#include "wav_io.h"
#endif

#ifndef MATRIX
#define MATRIX 1
#include "matrix.h"
#endif

#ifndef PERMUTATION
#define PERMUTATION 1
#include "permutation.h"
#endif

#define DIM 2

void usage(char *s)
{
  printf("%s -{options} <infile> {-o1 <out1file>} {-o2 <out1file>}\n",s);
  printf(" <infile> must be *.wav\n\n");
  printf(" options\n");
  printf("   -N     <int>  : NFFT, number for FFT,\n");
  printf("                   default=64,    This must be power of 2.\n");
  printf("   -T     <int>  : TAU,  number for making filter,\n");
  printf("                   default=NFFT/2,0 < TAU  < NFFT.\n");
  printf("   -S     <int>  : Time shift for spectrogram,\n");
  printf("                   default=20,    0 < SHIFT< NFFT.\n");
  printf("   -M     <real> : Time for Moving Average,\n");
  printf("                   default=0.04[s].\n");
  printf("   -m     <int>  : Number of matrixes to be diagonalized.\n");
  printf("                   default=40.\n");
  printf("   -P     <real> : Time for Solving Permutation,\n");
  printf("                   default=1.0[s].\n");
  printf("   -h            : print this message.\n\n");
  printf("   -o  <out_head>: Head part of the output file,\n");
  printf("                   default=<infile>, <out_head>_sep{1,2}.wav\n");
  exit(1);
}
      
int main(int argc, char *argv[])
{
  char 
    infile[1024],
    outfile[2][1024],
    append1[]      = "_sep1.wav",
    append2[]      = "_sep2.wav";
  int len,slen,i,j,k,l,f,t,FS,data4perm,ma,NFFT=512,TAU=0,
    SHIFT=20,MNUM=40;
  short sep_x[2];
  SCALAR MA=0.04,T4PERM=1.0,tmp,tmp_x;
  wavfmt fmt;
  vector window,xf_r,xf_i;
  matrix x;
  array3 Wsep_r,Wsep_i,Winv_r,Winv_i,sph_r,sph_i,rot_r,rot_i;
  array4 sp_r,sp_i,fil_t,Corr_r, Corr_i;
  FILE *fp;

  /* options */
  if(argc<2) usage(argv[0]);

  i=1;
  while(i<argc && argv[i][0]=='-'){
    switch(argv[i][1]){
    case 'N':
      ++i; NFFT   = atoi(argv[i]); break;
    case 'T':
      ++i; TAU    = atoi(argv[i]); break;
    case 'S':
      ++i; SHIFT  = atoi(argv[i]); break;
    case 'M':
      ++i; MA     = (SCALAR)atof(argv[i]); break;
    case 'm':
      ++i; MNUM   = atoi(argv[i]); break;
    case 'P':
      ++i; T4PERM = (SCALAR)atof(argv[i]); break;
    case 'h':
      usage(argv[0]); break;
    }
    ++i;
  }

  if(NFFT <= 0    || 
     (log10((double)NFFT)/log10(2.0)-floor(log10((double)NFFT)/log10(2.0)))!=0)
    usage(argv[0]);
  if(TAU   > NFFT || TAU   < 0) usage(argv[0]);
  if(SHIFT > NFFT || SHIFT < 0) usage(argv[0]);

  if(TAU==0)TAU=NFFT/2;

  if(i==argc)usage(argv[0]);
  else if(argc==(i+1)){
    strcpy(infile, argv[i]);
    strncpy(outfile[0], infile, strlen(infile)-4);
    strncpy(outfile[1], infile, strlen(infile)-4);
    outfile[0][strlen(infile)-4]='\0';
    outfile[1][strlen(infile)-4]='\0';
    strcat(outfile[0], append1);
    strcat(outfile[1], append2);
  }
  else if(argc>(i+1)){
    strcpy(infile, argv[i]);
    if(strcmp(argv[i+1],"-o")==0){
      strcpy(outfile[0], argv[i+2]);
      strcpy(outfile[1], argv[i+2]);
      outfile[0][strlen(argv[i+2])]='\0';
      outfile[1][strlen(argv[i+2])]='\0';
      strcat(outfile[0], append1);
      strcat(outfile[1], append2);
    }
    else usage(argv[0]);
  }
  
  /* options */

  len=read_wav(infile, &fmt);

  x = new_matrix(DIM,(int)len+NFFT-1);

  WaveData_2_DoubleMat(fmt, x);

  for(t=0;t<NFFT-1;++t)for(k=0;k<DIM;++k)x[k][len+t]=0;

  FS=fmt.FS;

  data4perm=((int)(T4PERM*FS))-(((int)(T4PERM*FS))-NFFT)%((int)SHIFT);
  if(data4perm>len)data4perm=len-(len-NFFT)%((int)SHIFT);
  slen=spl(data4perm, NFFT, SHIFT);
  ma=(int)floor((MA*FS)/((double)SHIFT));

  window = new_vector(NFFT); hamming(NFFT, window);
  xf_r   = new_vector(NFFT);
  xf_i   = new_vector(NFFT);

  Wsep_r = new_array3(NFFT/2+1,DIM,DIM);
  Wsep_i = new_array3(NFFT/2+1,DIM,DIM);
  Winv_r = new_array3(NFFT/2+1,DIM,DIM);
  Winv_i = new_array3(NFFT/2+1,DIM,DIM);

  for(f=0;f<NFFT/2+1;++f){
    eye(DIM, Wsep_r[f]); zeros(DIM, DIM, Wsep_i[f]);
    eye(DIM, Winv_r[f]); zeros(DIM, DIM, Winv_i[f]);
  }

  fil_t  = new_array4(DIM,DIM,DIM,NFFT);

  Corr_r = new_array4(MNUM,NFFT/2+1,DIM,DIM);
  Corr_i = new_array4(MNUM,NFFT/2+1,DIM,DIM);
  corr_f(DIM, NFFT, SHIFT, len, MNUM, window, x, Corr_r, Corr_i);

  sph_r  = new_array3(NFFT/2+1,DIM,DIM);
  sph_i  = new_array3(NFFT/2+1,DIM,DIM);
  sphering_c(DIM, NFFT, MNUM, sph_r, sph_i, Corr_r, Corr_i);

  rot_r  = new_array3(NFFT/2+1,DIM,DIM);
  rot_i  = new_array3(NFFT/2+1,DIM,DIM);
  joint_diag_f(DIM, MNUM, NFFT, Corr_r, Corr_i, rot_r, rot_i);

  free_array4(Corr_r,MNUM,NFFT/2+1); free_array4(Corr_i,MNUM,NFFT/2+1);

  sph_rot2Wsep(DIM, NFFT, sph_r, sph_i, rot_r, rot_i, Wsep_r, Wsep_i);

  free_array3(sph_r,NFFT/2+1); free_array3(sph_i,NFFT/2+1);
  free_array3(rot_r,NFFT/2+1); free_array3(rot_i,NFFT/2+1);

  for(f=0;f<NFFT/2+1;++f)
    matinv_complex(DIM,Wsep_r[f],Wsep_i[f],Winv_r[f],Winv_i[f]);

  sp_r   = new_array4(DIM,DIM,NFFT/2+1,slen);
  sp_i   = new_array4(DIM,DIM,NFFT/2+1,slen);

  for(k=0;k<DIM;++k)for(i=0;i<DIM;++i)
    for(f=0;f<NFFT/2+1;++f)for(t=0;t<slen;++t){
	  sp_r[k][i][f][t]=0;
	  sp_i[k][i][f][t]=0;
    }

  for(t=0;t<slen;++t){
    for(j=0;j<DIM;++j){
      for(l=0;l<NFFT;++l){
	xf_r[l]=window[l]*x[j][t*SHIFT+l];
	xf_i[l]=0;
      }
      fft(NFFT,xf_r,xf_i);
      for(k=0;k<DIM;++k){
	for(f=0;f<NFFT/2+1;++f){
	  for(i=0;i<DIM;++i){
	    sp_r[k][i][f][t]+=
	      (Winv_r[f][i][k]*Wsep_r[f][k][j]
	       -Winv_i[f][i][k]*Wsep_i[f][k][j])*xf_r[f]
	      -(Winv_i[f][i][k]*Wsep_r[f][k][j]
		+Winv_r[f][i][k]*Wsep_i[f][k][j])*xf_i[f];
	    sp_i[k][i][f][t]+=
	      (Winv_r[f][i][k]*Wsep_r[f][k][j]
	       -Winv_i[f][i][k]*Wsep_i[f][k][j])*xf_i[f]
	      +(Winv_i[f][i][k]*Wsep_r[f][k][j]
		+Winv_r[f][i][k]*Wsep_i[f][k][j])*xf_r[f];
	  }
	}
      }
    }
  }

  solve_permutation(DIM,NFFT,ma,slen,sp_r,sp_i,Wsep_r,Wsep_i,Winv_r,Winv_i);

  free_array4(sp_r,DIM,DIM);  free_array4(sp_i,DIM,DIM);

  printf("Making filters.\n");
  W2fil_t(DIM,NFFT,TAU,window,Wsep_r,Wsep_i,Winv_r,Winv_i,fil_t);
  printf("Done.\n");

  free_array3(Wsep_r,NFFT/2+1); free_array3(Wsep_i,NFFT/2+1);
  free_array3(Winv_r,NFFT/2+1); free_array3(Winv_i,NFFT/2+1);

  for(k=0;k<DIM;++k){
    printf("Making %s\n",outfile[k]);
    fp=write_wav_head(outfile[k], &fmt); 
    for(t=0;t<len;++t){
      for(i=0;i<DIM;++i){
	tmp=0;
	for(j=0;j<DIM;++j){
	  conv(NFFT,NFFT,fil_t[k][i][j],&x[j][t],&tmp_x);
	  tmp+=tmp_x;
	}
	if(tmp>=1){
	  printf("data is too large and be clipped\n");
	  sep_x[i]=MAX-1;
	}
	else if(tmp<-1){
	  printf("data is too large and be clipped\n");
	  sep_x[i]=-MAX;
	}
	else sep_x[i]=(short)(MAX*tmp);
      }
      fwrite(sep_x,2,2,fp);
    }
    fclose(fp);
    printf("Done\n");
  }

  return(1);
}
