#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bss.h"
#include "spec.h"

#ifndef ARRAY
#define ARRAY 1
#include "array.h"
#endif

#ifndef WAV
#define WAV 1
#include "wav_io.h"
#endif

#ifndef MATRIX
#define MATRIX 1
#include "matrix.h"
#endif

#ifndef PERMUTATION
#define PERMUTATION 1
#include "permutation.h"
#endif

#define DIM 2
#define R NFFT

void usage(char *s)
{
  printf("%s -{options} <infile> {-o1 <out1file>} {-o2 <out1file>}\n",s);
  printf(" <infile> must be *.wav\n\n");
  printf(" options\n");
  printf("   -N     <int>  : NFFT, number for FFT,\n");
  printf("                   default=64,    This must be power of 2.\n");
  printf("   -T     <int>  : TAU,  number for making filter,\n");
  printf("                   default=NFFT/2,0 < TAU  < NFFT.\n");
  printf("   -S     <int>  : Time shift for spectrogram,\n");
  printf("                   default=20,    0 < SHIFT< NFFT.\n");
  printf("   -E     <real> : Learninf ration,\n");
  printf("                   default=0.02.\n");
  printf("   -M     <real> : Time for Moving Average,\n");
  printf("                   default=0.04[s].\n");
  printf("   -P     <real> : Time for Solving Permutation,\n");
  printf("                   default=1.0[s].\n");
  printf("   -h            : print this message.\n\n");
  printf("   -o  <out_head>: Head part of the output file,\n");
  printf("                   default=<infile>, <out_head>_sep{1,2}.wav\n");
  exit(1);
}
      
int main(int argc, char *argv[])
{
  char infile[1024],outfile[2][1024],
    append1[]="_sep1.wav",append2[]="_sep2.wav";
  int len,slen,i,j,k,l,f,t,FS,data4perm,t_s,ma,NFFT=64,TAU=0,
    SHIFT=20;
  SCALAR tmp,ETA=0.02,MA=0.04,T4PERM=1.0;
  wavfmt fmt;
  vector window;
  matrix x,xf_r,xf_i;
  array3 sep_x,Wsep_r,Wsep_i,Winv_r,Winv_i;
  array4 sp_r,sp_i,fil_r,fil_i,fil_t;


  /* options */
  if(argc<2){
    usage(argv[0]);
  }
  i=1;
  while(i<argc && argv[i][0]=='-'){
    switch(argv[i][1]){
    case 'N':
      ++i; NFFT=atoi(argv[i]); break;
    case 'T':
      ++i; TAU=atoi(argv[i]);  break;
    case 'S':
      ++i; SHIFT=atoi(argv[i]);  break;
    case 'E':
      ++i; ETA=(SCALAR)atof(argv[i]);  break;
    case 'M':
      ++i; MA =(SCALAR)atof(argv[i]);  break;
    case 'P':
      ++i; T4PERM =(SCALAR)atof(argv[i]);  break;
    case 'h':
      usage(argv[0]); break;
    }
    ++i;
  }

  if(NFFT <= 0    || 
     (log10((double)NFFT)/log10(2.0)-floor(log10((double)NFFT)/log10(2.0)))!=0)
    usage(argv[0]);
  if(TAU   > NFFT || TAU   < 0) usage(argv[0]);
  if(SHIFT > NFFT || SHIFT < 0) usage(argv[0]);

  if(TAU==0)TAU=NFFT/2;

  if(i==argc)usage(argv[0]);
  else if(argc==(i+1)){
    strcpy(infile, argv[i]);
    strncpy(outfile[0], infile, strlen(infile)-4);
    strncpy(outfile[1], infile, strlen(infile)-4);
    outfile[0][strlen(infile)-4]='\n';
    outfile[1][strlen(infile)-4]='\n';
    strcat(outfile[0], append1);
    strcat(outfile[1], append2);
  }
  else if(argc>(i+1)){
    strcpy(infile, argv[i]);
    if(strcmp(argv[i+1],"-o")==0){
      strcpy(outfile[0], argv[i+2]);
      strcpy(outfile[1], argv[i+2]);
      outfile[0][strlen(argv[i+2])]='\n';
      outfile[1][strlen(argv[i+2])]='\n';
      strcat(outfile[0], append1);
      strcat(outfile[1], append2);
    }
    else usage(argv[0]);
  }

  /* options */

  len=read_wav(infile, &fmt);

  x=new_matrix(DIM,(int)len+NFFT/2);

  WaveData_2_DoubleMat(fmt, x);

  for(t=0;t<NFFT/2;++t)for(k=0;k<DIM;++k)x[k][len+t]=0;

  FS=fmt.FS;

  data4perm=((int)(T4PERM*FS))-(((int)(T4PERM*FS))-NFFT)%((int)SHIFT);
  if(data4perm>len)data4perm=len-(len-NFFT)%((int)SHIFT);
  slen=spl(data4perm, NFFT, SHIFT);
  ma=(int)floor((MA*FS)/((SCALAR)SHIFT));

  /* printf("read data end\n"); */

  window = new_vector(NFFT); hamming(NFFT, window);
  xf_r   = new_matrix(DIM,NFFT);
  xf_i   = new_matrix(DIM,NFFT);
  sep_x  = new_array3(DIM,DIM,len);
  Wsep_r = new_array3(NFFT/2+1,DIM,DIM);
  Wsep_i = new_array3(NFFT/2+1,DIM,DIM);
  Winv_r = new_array3(NFFT/2+1,DIM,DIM);
  Winv_i = new_array3(NFFT/2+1,DIM,DIM);
  for(f=0;f<NFFT/2+1;++f){
    eye(DIM, Wsep_r[f]); zeros(DIM, DIM, Wsep_i[f]);
    eye(DIM, Winv_r[f]); zeros(DIM, DIM, Winv_i[f]);
  }
  fil_r  = new_array4(DIM,DIM,DIM,(NFFT/2+1));
  fil_i  = new_array4(DIM,DIM,DIM,(NFFT/2+1));
  fil_t  = new_array4(DIM,DIM,DIM,NFFT);
  sp_r   = new_array4(DIM,DIM,NFFT/2+1,slen);
  sp_i   = new_array4(DIM,DIM,NFFT/2+1,slen);

  for(t=0;t<data4perm-NFFT+1;++t){
    if(t>0 && (t*10)%((int)FS)==0){printf(".");fflush(stdout);}
    if(t>0 && t%((int)FS)==0)printf("\n");

    for(i=0;i<DIM;++i){
      for(l=0;l<NFFT;++l){
	xf_r[i][l]=window[l]*x[i][t+l];
	xf_i[i][l]=0;
      }
      fft(NFFT,xf_r[i],xf_i[i]);
    }
    update_W(DIM,NFFT,ETA,xf_r,xf_i,Wsep_r,Wsep_i,Winv_r,Winv_i,R);

    if((t%((int)SHIFT))==0){
      t_s=(int)(((SCALAR)t)/((SCALAR)SHIFT));
      W2fil_f(DIM,NFFT,TAU,Wsep_r,Wsep_i,Winv_r,Winv_i,fil_r,fil_i);
      for(k=0;k<DIM;++k){
	for(f=0;f<NFFT/2+1;++f){
	  for(i=0;i<DIM;++i){
	    for(sp_r[k][i][f][t_s]=0,sp_i[k][i][f][t_s]=0,j=0;j<DIM;++j){
	      sp_r[k][i][f][t_s]+=
		fil_r[k][i][j][f]*xf_r[j][f]-fil_i[k][i][j][f]*xf_i[j][f];
	      sp_i[k][i][f][t_s]+=
		fil_r[k][i][j][f]*xf_i[j][f]+fil_i[k][i][j][f]*xf_r[j][f];
	    }
	  }
	}
      }
    }
  }

  free_array4(fil_r,DIM,DIM); free_array4(fil_i,DIM,DIM);

  solve_permutation(DIM,NFFT,ma,slen,sp_r,sp_i,Wsep_r,Wsep_i,Winv_r,Winv_i);
  for(k=0;k<DIM;++k)for(i=0;i<DIM;++i)
    ispecgram(sep_x[k][i],slen,data4perm,SHIFT,NFFT,window,
	      sp_r[k][i],sp_i[k][i]);

  free_array4(sp_r,DIM,DIM);  free_array4(sp_i,DIM,DIM);

  /* printf("preparation end\n"); */

  for(t=data4perm-NFFT+1;t<len-NFFT/2;++t){
    if(t>data4perm && ((t-data4perm)*10)%((int)FS)==0){
      printf(".");fflush(stdout);
    }
    if(t>data4perm && (t-data4perm)%((int)FS)==0)printf("\n");

    for(j=0;j<DIM;++j){
      for(l=0;l<NFFT;++l){xf_r[j][l]=window[l]*x[j][t+l];xf_i[j][l]=0;}
      fft(NFFT,xf_r[j],xf_i[j]);
    }

    update_W(DIM,NFFT,ETA,xf_r,xf_i,Wsep_r,Wsep_i,Winv_r,Winv_i,R/2);
    W2fil_t(DIM,NFFT,TAU,window,Wsep_r,Wsep_i,Winv_r,Winv_i,fil_t);
      
    for(k=0;k<DIM;++k)for(i=0;i<DIM;++i)
      for(sep_x[k][i][t]=0,j=0;j<DIM;++j){
	conv(NFFT,NFFT,fil_t[k][i][j],&x[j][t],&tmp);
	sep_x[k][i][t]+=tmp;
      }
  }

  printf("\n");

  for(i=0;i<DIM;++i){
    printf("Making %s\n",outfile[i]);
    DoubleMat_2_WaveData(len,sep_x[i],&fmt);write_wav(outfile[i], &fmt);
  }

  return(1);
}
