#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "matrix.h"

void error(char *message)
{
  printf("\n%s\n",message);
  exit(0);
}

vector newvec(int n)
{
  return (SCALAR *)malloc(sizeof(SCALAR)*n);
}

matrix newmat(int nrow, int ncol)
{
  int i;
  matrix a;
  
  a=(SCALAR **)malloc((nrow+1)*sizeof(SCALAR));
  if(a==NULL) return NULL;
  for(i=0;i<nrow;++i){
    a[i]=(SCALAR *)malloc(sizeof(SCALAR)*ncol);
    if(a[i]==NULL){
      while(--i>=0)free(a[i]);
      free(a);return(NULL);
    }
  }
  a[nrow]=NULL;
  return a;
}

vector new_vector(int n)
{
  vector v;
  
  v=newvec(n);
  if(v==NULL)error("No Memory Space\n");
  return(v);
}

matrix new_matrix(int nrow, int ncol)
{
  matrix a;

  /*printf("Request is to malloc %ldbyte.\n", nrow*ncol*sizeof(SCALAR));*/
  a=newmat(nrow, ncol);
  if(a==NULL) error("No Space\n");
  return a;
}

void free_vector(vector v)
{
  free(v);
}

void free_matrix(matrix a)
{
  matrix b;
  
  b=a;
  while(*b!=NULL) free(*b++);
  free(a);
}

void zeros(int n, int m, matrix a)
{
  int i,j;

  for(i=0;i<n;++i)for(j=0;j<m;++j)
    a[i][j]=0;
}

void eye(int n, matrix a)
{
  int i;
  zeros(n, n, a);
  for(i=0;i<n;++i)a[i][i]=1;
}

void copy_mat(int row, int col, matrix a, matrix c)
{
  int i,j;

  for(i=0;i<row;++i)for(j=0;j<col;++j)
    c[i][j]=a[i][j];
}

void printmat(int nrow, int mcol, matrix a)
{
  int i,j;
  SCALAR tmp1;

  printf("Size is %d %d\n",nrow,mcol);
  for(i=0;i<nrow;++i){
    for(j=0;j<mcol;++j){
      tmp1=a[i][j];
      if(fabs(tmp1)<EPS_HYOUJI)tmp1=0;
      printf("%g ",tmp1);
    }
    printf("\n");
  }
}

void printvec(vector v, int n)
{
  int j;
  for(j=0;j<n;j++)printf("%g ",v[j]);
  printf("\n");
}

SCALAR innerproduct(int n, vector u, vector v)
{
  int i,n5;
  SCALAR s;

  s=0; n5=n%5;
  for(i=0;i<n5;i++) s+=u[i]*v[i];
  for(i=n5;i<n;i+=5)
    s+=u[i]*v[i]+u[i+1]*v[i+1]+u[i+2]*v[i+2]
      +u[i+3]*v[i+3]+u[i+4]*v[i+4];
  return s;
}

SCALAR powvec(int dim, vector data)
{
  int i;
  double poi=0;

  for(i=0;i<dim;++i)poi+=data[i]*data[i];
  poi=(SCALAR)sqrt(((double)poi));
  return(poi);
}

void trans_mat(int n, int m, matrix a, matrix c)
{
  int i,j;
  for(i=0;i<n;++i)for(j=0;j<m;++j)
    c[j][i]=a[i][j];
}

void sum_mat(int n, int m, matrix a, matrix b, matrix c)
{
  int i,j;

  for(i=0;i<n;++i)
    for(j=0;j<m;++j)
      c[i][j]=a[i][j]+b[i][j];
}

void sub_mat(int n, int m, matrix a, matrix b, matrix c)
{
  int i,j;

  for(i=0;i<n;++i)
    for(j=0;j<m;++j)
      c[i][j]=a[i][j]-b[i][j];
}

void multiply(int n, int l, int m, matrix a, matrix b, matrix c)
{
  int i,j,k;
  SCALAR s;

  for(i=0; i<n; i++)
    for(j=0; j<m; j++){
      s=0;
      for(k=0;k<l;k++)s+=a[i][k]*b[k][j];
      c[i][j]=s;
    }
}

void mult_cmp(int n, int l, int m, 
	       matrix a_r, matrix a_i, 
	       matrix b_r, matrix b_i,
	       matrix c_r, matrix c_i,
	       matrix tmp_r, matrix tmp_i)
{
  multiply(n, l, m, a_r, b_r, tmp_r);
  multiply(n, l, m, a_i, b_i, tmp_i);

  sub_mat(n, m, tmp_r, tmp_i, c_r);

  multiply(n, l, m, a_i, b_r, tmp_r);
  multiply(n, l, m, a_r, b_i, tmp_i);

  sum_mat(n, m, tmp_r, tmp_i, c_i);

}

void mult2(int n, matrix a, matrix b)
{
  int i,j,k;
  SCALAR s;
  vector tmp;
  
  tmp=new_vector(n);
  
  for(i=0;i<n;i++){
    for(k=0;k<n;k++) tmp[k]=a[i][k];
    for(j=0;j<n;j++){
      s=0;
      for(k=0;k<n;k++) s+=tmp[k]*b[k][j];
      a[i][j]=s;
    }
  }
  free_vector(tmp);
}

SCALAR lu(int n, matrix a, int *ip)
{
  int i,j,k,ii,ik;
  SCALAR t,u,det;
  vector weight;
  
  weight = new_vector(n);
  det = 0;
  for(k=0;k<n;k++){
    ip[k]=k;
    u=0;
    for(j=0;j<n;j++){
      t=(SCALAR)fabs((double)(a[k][j])); if(t>u) u=t;
    }
    if(u==0) goto EXIT;
    weight[k]=1/u;
  }
  det=1;
  for(k=0;k<n;k++){
    u=-1;
    for(j=k,i=k;i<n;i++){
      ii=ip[i];
      t=(SCALAR)fabs((double)a[ii][k])*weight[ii];
      if(t>u){u=t;j=i;}
    }
    ik=ip[j];
    if(j!=k){
      ip[j]=ip[k]; ip[k]=ik;
      det=-det;
    }
    u=a[ik][k];det*=u;
    if(u==0) goto EXIT;
    for(i=k+1;i<n;i++){
      ii=ip[i];
      t=(a[ii][k]/=u);
      for(j=k+1;j<n;j++)
	a[ii][j]-=t*a[ik][j];
    }
  }
 EXIT:
  free_vector(weight);
  return(det);
}

SCALAR matinv(int n, matrix a, matrix a_inv)
{
  int i,j,k,ii;
  SCALAR t, det;
  int *ip;

  ip=(int *)malloc(sizeof(int)*n);
  if(ip==NULL) error("No Memory Space");
  det=lu(n,a,ip);
  if(det !=0)
    for(k=0;k<n;k++){
      for(i=0;i<n;i++){
	ii=ip[i]; t=(ii==k);
	for(j=0;j<i;j++)
	  t-=a[ii][j]*a_inv[j][k];
	a_inv[i][k]=t;
      }
      for(i=n-1;i>=0;i--){
	t=a_inv[i][k]; ii=ip[i]; 
	for(j=i+1;j<n;j++)
	  t-=a[ii][j]*a_inv[j][k];
	a_inv[i][k]=t/a[ii][i];
      }
    }
  free(ip);
  return(det);
}

SCALAR matinv_complex(int dim,
		      matrix in_r, matrix in_i,
		      matrix out_r,matrix out_i)
{
  static int last_dim=0;
  static matrix tmpmat=NULL,invmat=NULL;
  int i,j;
  SCALAR det;
  
  if(dim!=last_dim){
    if(tmpmat!=NULL) free_matrix(tmpmat);
    if(invmat!=NULL) free_matrix(invmat);
    tmpmat=new_matrix(2*dim,2*dim);
    invmat=new_matrix(2*dim,2*dim);
    last_dim=dim;
  }

  for(i=0;i<dim;++i)
    for(j=0;j<dim;++j){
      tmpmat[i][j]        = in_r[i][j];
      tmpmat[i+dim][j+dim]= in_r[i][j];
      tmpmat[i][j+dim]    =-in_i[i][j];
      tmpmat[i+dim][j]    = in_i[i][j];
    }
  if((det=matinv(2*dim,tmpmat,invmat))!=0){
    for(i=0;i<dim;++i)
      for(j=0;j<dim;++j){
	out_r[i][j]= invmat[i][j];
	out_i[i][j]= invmat[i+dim][j];
      }
  }
  return(det);
}

SCALAR matinv2(int n, matrix aa, matrix a_inv)
{
  int i,j,k,ii;
  SCALAR t, det;
  int *ip;
  matrix a;

  a=new_matrix(n,n);
  copy_mat(n,n,aa,a);

  ip=(int *)malloc(sizeof(int)*n);
  if(ip==NULL) error("No Memory Space");
  det=lu(n,a,ip);
  if(det !=0)
    for(k=0;k<n;k++){
      for(i=0;i<n;i++){
	ii=ip[i]; t=(ii==k);
	for(j=0;j<i;j++)
	  t-=a[ii][j]*a_inv[j][k];
	a_inv[i][k]=t;
      }
      for(i=n-1;i>=0;i--){
	t=a_inv[i][k]; ii=ip[i]; 
	for(j=i+1;j<n;j++)
	  t-=a[ii][j]*a_inv[j][k];
	a_inv[i][k]=t/a[ii][i];
      }
    }
  free(ip);
  free_matrix(a);
  return(det);
}

SCALAR eig2(SCALAR a, SCALAR b, SCALAR c, SCALAR *x, SCALAR *y)
     /* return x y, which corresopond to the eigen vector for the
       largest aigenvector of 2x2 simmetric matirx [a,c;c,b] */
{
  SCALAR lambda,tmp,tmp1;

  tmp1=a-b;
  
  lambda=0.5*(a+b+(SCALAR)sqrt((double)(tmp1*tmp1+4*c*c)));

  tmp1=(a-lambda);

  tmp=(SCALAR)sqrt((double)(tmp1*tmp1+c*c));

  if(c<0){
    *x=-c/tmp;
    *y=tmp1/tmp;
  }
  else{
    *x=c/tmp;
    *y=-tmp1/tmp;
  }

  return lambda;
}

SCALAR house(int n, vector x)
{
  int i;
  SCALAR s,t;

  s=(SCALAR)sqrt((double)(innerproduct(n,x,x)));

  if(s!=0){
    if(x[0]<0)s=-s;
    x[0]+=s; t=(SCALAR)(1/sqrt((double)(x[0]*s)));
    for(i=0;i<n;i++) x[i]*=t;
  }
  return(-s);
}

void tridiagonalize(int n, matrix a, vector d, vector e)
{
  int i,j,k;
  SCALAR s,t,p,q;
  vector v,w;

  for(k=0;k<n-2;k++){
    v=a[k]; d[k]=v[k];
    e[k]=house(n-k-1,&v[k+1]);
    if(e[k] == 0) continue;
    for(i=k+1;i<n;i++){
      s=0;
      for(j=k+1;j<i;j++) s+=a[j][i]*v[j];
      for(j=i;  j<n;j++) s+=a[i][j]*v[j];
      d[i]=s;
    }
    t=innerproduct(n-k-1, &v[k+1], &d[k+1])/2;
    
    for(i=n-1;i>k; i--){
      p=v[i];q=d[i]-t*p; d[i]=q;
      for(j=i;j<n;j++)
	a[i][j]-=p*d[j]+q*v[j];
    }
  }
  if(n>=2){ 
    d[n-2]=a[n-2][n-2];
    e[n-2]=a[n-2][n-1];
  }
  if(n>=1)
    d[n-1]=a[n-1][n-1];
  for(k=n-1;k>=0;k--){
    v=a[k];
    if(k<n-2){
      for(i=k+1;i<n;i++){
	w=a[i];
	t=innerproduct(n-k-1, &v[k+1],&w[k+1]);
	for(j=k+1;j<n;j++)
	  w[j]-=t*v[j];
      }
    }
    for(i=0;i<n;i++) v[i]=0;
    v[k]=1;
  }
}

SCALAR eig3(matrix a, SCALAR *x, SCALAR *y, SCALAR *z)
{
  int i,j,k,h,iter;
  SCALAR c,s,t,w,xx,yy;
  vector v,d,e;
  
  d=new_vector(3);
  e=new_vector(3);

  tridiagonalize(3,a,d,&e[1]);
  e[0]=0;

  for(h=2;h>0;h--){
    j=h;

    while( fabs(e[j]) > (EPS *(fabs(d[j-1])+fabs(d[j]))) ){
      j--;
      if(j<=0) break;
    }
    if(j==h) continue;
    iter = 0;

    do {
      if(++iter > MAX_ITER){
	printf("Cannot get eigenvalue\n");
	exit(0);
      }
      w=(d[h-1]-d[h])/2;
      t=e[h]*e[h];
      s=(SCALAR)sqrt((double)(w*w+t)); if(w<0)s=-s;
      xx=d[j]-d[h]+t/(w+s); yy=e[j+1];

      for(k=j;k<h;k++){
	if(fabs(xx)>=fabs(yy)){
	  t=-yy/xx; c=(SCALAR)(1/sqrt(t*t+1));
	  s=t*c;
	}
	else {
	  t=-xx/yy; s=(SCALAR)(1/sqrt(t*t+1));
	  c=t*s;
	}
	w=d[k]-d[k+1];
	t=(w*s+2*c*e[k+1])*s;
	d[k]-=t; d[k+1]+=t;
	if(k>j)e[k]=c*e[k]-s*yy;
	e[k+1]+=s*(c*w-2*s*e[k+1]);

	for(i=0;i<3;i++){
	  xx=a[k][i]; yy=a[k+1][i];
	  a[k  ][i]=c*xx-s*yy;
	  a[k+1][i]=s*xx+c*yy;
	}
	if(k<h-1){
	  xx=e[k+1];yy=-s*e[k+2];
	  e[k+2]*=c;
	}
      }
    }while(fabs(e[h])>EPS*(fabs(d[h-1])+fabs(d[h])));
  }
  
  for(k=0;k<2;k++){
    h=k; t=d[h];
    for(i=k+1;i<3;i++)
      if(d[i]>t){h=i;t=d[h];}
    d[h]=d[k];d[k]=t;
    v=a[h];a[h]=a[k];a[k]=v;
    }

  *x=a[0][0];
  *y=a[0][1];
  *z=a[0][2];
  return(d[0]);
}


/*
  n x n matrix, 
  a's engenvectors will be stored in a as a'. 
  d is eigen values, e is working space.
*/
void eigen(int n, matrix a, vector d, vector e)
{
  int i,j,k,h,iter;
  SCALAR c,s,t,w,x,y;
  vector v;

  tridiagonalize(n,a,d,&e[1]);
  e[0]=0;

  for(h=n-1;h>0;h--){
    j=h;

    while( fabs(e[j]) > (EPS *(fabs(d[j-1])+fabs(d[j]))) ){
      j--;
      if(j<=0) break;
    }
    if(j==h) continue;
    iter = 0;

    do {
      if(++iter > MAX_ITER){
	printf("Cannot get eigenvalue\n");
	exit(0);
      }
      w=(d[h-1]-d[h])/2;
      t=e[h]*e[h];
      s=(SCALAR)sqrt((double)(w*w+t)); if(w<0)s=-s;
      x=d[j]-d[h]+t/(w+s); y=e[j+1];

      for(k=j;k<h;k++){
	if(fabs(x)>=fabs(y)){
	  t=-y/x; c=(SCALAR)(1/sqrt((double)(t*t+1)));
	  s=t*c;
	}
	else {
	  t=-x/y; s=(SCALAR)(1/sqrt((double)(t*t+1)));
	  c=t*s;
	}
	w=d[k]-d[k+1];
	t=(w*s+2*c*e[k+1])*s;
	d[k]-=t; d[k+1]+=t;
	if(k>j)e[k]=c*e[k]-s*y;
	e[k+1]+=s*(c*w-2*s*e[k+1]);

	for(i=0;i<n;i++){
	  x=a[k][i]; y=a[k+1][i];
	  a[k  ][i]=c*x-s*y;
	  a[k+1][i]=s*x+c*y;
	}
	if(k<h-1){
	  x=e[k+1];y=-s*e[k+2];
	  e[k+2]*=c;
	}
      }
    }while(fabs(e[h])>EPS*(fabs(d[h-1])+fabs(d[h])));
  }
  
  for(k=0;k<n-1;k++){
    h=k; t=d[h];
    for(i=k+1;i<n;i++)
      if(d[i]>t){h=i;t=d[h];}
    d[h]=d[k];d[k]=t;
    v=a[h];a[h]=a[k];a[k]=v;
    }
}


