#include <stdio.h>
#include "permutation.h"

void sorting(int dim, vector data, int index[])
{
  int i,j,k,xx;
  SCALAR x;

  for(i=0;i<dim;++i)index[i]=i;

  k = dim-1;
  while(k>=0){
    j=-1;
    for(i=1; i<=k; i++){
      if(data[i-1] > data[i]){
	j=i-1;
	x=data[j];   data[j]=data[i];   data[i]=x;
	xx=index[j]; index[j]=index[i]; index[i]=xx;
      }
    }
    k=j;
  }
}

void permutation(int N, int pos, int k, char *ok, int *p, 
		 matrix a, int *p_max, SCALAR *max)
{
  int i;
  SCALAR tmp_max;

  p[pos] = k;
  if (pos != N - 1){
    ok[k] = 0;
    for (i = 1; i <= N; i++)
      if (ok[i]) permutation(N,pos + 1, i, ok, p, a, p_max, max);
    ok[k] = 1;
  }
  else{
    for(tmp_max=0,i=0;i<N;++i){
      tmp_max+=a[i][p[i]-1];
    }
    if(tmp_max>*max){
      *max=tmp_max;
      for(tmp_max=0,i=0;i<N;++i)p_max[i]=p[i];
    }
  }
}

SCALAR solveperm(int *count, int N, char *ok, int *p, matrix a, int *p_max)
{
  int k;
  SCALAR max=0;
  
  *count = 0;
  for (k = 1; k <= N; k++) ok[k] = 1;
  for (k = 1; k <= N; k++) permutation(N, 0, k, ok, p, a, p_max, &max);
  return(max);
}

void solve_permutation(int dim, int nfft, int ma, int slen, 
		       array4 sp_r, array4 sp_i, 
		       array3 Wsep_r, array3 Wsep_i,
		       array3 Winv_r, array3 Winv_i)
{
  char *ok;
  int i,j,k,kk,f,t,*index,*p_perm,tmp_int, *perm;
  SCALAR **pivo,max;
  vector tt,ma_fil,order,p_env;
  matrix env,in_pro,pow;
  array3 spgm;

  printf("\nSolving Permutation\n"); 

  index = malloc(sizeof(int)*(nfft/2+1));
  p_perm= malloc(sizeof(int)*dim);
  perm  = malloc(sizeof(int)*dim);
  ok    = malloc(sizeof(char)*(dim+1));
  pivo  = (SCALAR **)malloc(sizeof(SCALAR)*dim);
  tt    = new_vector(slen);
  ma_fil= new_vector(ma);
  order = new_vector(nfft/2+1);
  p_env = new_vector(dim);
  env   = new_matrix(dim,slen-ma+1);
  in_pro= new_matrix(dim,dim);
  pow   = new_matrix(dim,nfft/2+1);
  spgm  = new_array3(dim,nfft/2+1,slen-ma+1);

  for(i=0;i<ma;++i)ma_fil[i]=(SCALAR)1/((SCALAR)ma);

  for(k=0;k<dim;++k){
    for(f=0;f<nfft/2+1;++f){
      for(t=0;t<slen;++t){
	for(tt[t]=0,i=0;i<dim;++i){
	  tt[t]+=(SCALAR)sqrt((double)(sp_r[k][i][f][t]*sp_r[k][i][f][t]
				       +sp_i[k][i][f][t]*sp_i[k][i][f][t]));
	}
      }
      conv(ma,slen,ma_fil,tt,spgm[k][f]);
      pow[k][f]=powvec(slen-ma+1,spgm[k][f]);
    }
  }
  free_vector(tt);

  for(f=0;f<nfft/2+1;++f){
    for(i=0;i<dim-1;++i){
      for(order[f]=0,j=i+1;j<dim;++j){
	order[f]+=innerproduct(slen-ma+1,spgm[i][f],spgm[j][f])
	  /(pow[i][f]*pow[j][f]);
      }
    }
  }

  sorting(nfft/2+1, order, index);

  for(i=0;i<dim;++i){
    for(t=0;t<slen-ma+1;++t){
      env[i][t]=spgm[i][index[0]][t];
    }
    p_env[i]=pow[i][index[0]];
  }

  for(f=1;f<nfft/2+1;++f){
    for(i=0;i<dim;++i){
      for(j=0;j<dim;++j){
	in_pro[i][j]=innerproduct(slen-ma+1,env[i],spgm[j][index[f]])
	  /(p_env[i]*pow[j][index[f]]);
      }
    }
    max=solveperm(&tmp_int, dim, ok, p_perm, in_pro, perm);
    for(k=0;k<dim;++k)if(perm[k]!=(k+1)){
      printf("in Frequency %d\n",index[f]);
      for(i=0;i<dim;++i){
	for(kk=0;kk<dim;++kk)pivo[kk]=sp_r[perm[kk]-1][i][index[f]];
	for(kk=0;kk<dim;++kk)sp_r[kk][i][index[f]]=pivo[kk];
      }

      for(i=0;i<dim;++i){
	for(kk=0;kk<dim;++kk)pivo[kk]=sp_i[perm[kk]-1][i][index[f]];
	for(kk=0;kk<dim;++kk)sp_i[kk][i][index[f]]=pivo[kk];
      }

      for(kk=0;kk<dim;++kk)pivo[kk]=Wsep_r[index[f]][perm[kk]-1];
      for(kk=0;kk<dim;++kk)Wsep_r[index[f]][kk]=pivo[kk];
      
      for(kk=0;kk<dim;++kk)pivo[kk]=Wsep_i[index[f]][perm[kk]-1];
      for(kk=0;kk<dim;++kk)Wsep_i[index[f]][kk]=pivo[kk];
	
      matinv_complex(dim,Wsep_r[index[f]],Wsep_i[index[f]],
		     Winv_r[index[f]],Winv_i[index[f]]);
	
      for(i=0;i<dim;++i){
	for(t=0;t<slen-ma+1;++t)
	  env[i][t]+=max*spgm[perm[i]-1][index[f]][t]/((SCALAR)dim);
	p_env[i]=powvec(slen-ma+1,env[i]);
      }
      break;
    }

    for(i=0;i<dim;++i){
      for(t=0;t<slen-ma+1;++t)env[i][t]+=spgm[i][index[f]][t]/pow[i][index[f]];
      p_env[i]=powvec(slen-ma+1,env[i]);
    }
  }

  free(index); free(p_perm); free(perm); free(ok); free(pivo);
  free_vector(ma_fil); free_vector(order); free_vector(p_env);
  free_matrix(env); free_matrix(in_pro); free_matrix(pow);
  free_array3(spgm,dim);

  printf("Done\n"); 
}

