#include "spec.h"

static void make_sintbl(int n, vector sintbl)
{
  int i,n2,n4,n8;
  SCALAR c,s,dc,ds,t;

  n2=n/2; n4=n/4; n8=n/8;
  t =sin(PI/n);
  dc=2*t*t; ds=sqrt(dc*(2-dc));
  t =2*dc;  c =sintbl[n4]=1;  s=sintbl[0]=0;
  for(i=1;i<n8;i++){
    c-=dc; dc+=t*c;
    s+=ds; ds-=t*s;
    sintbl[i]=s; sintbl[n4-i]=c;
  }
  if(n8!=0) sintbl[n8]=sqrt(0.5);
  for(i=0;i<n4;i++)
    sintbl[n2-i]=sintbl[i];
  for(i=0;i<n2+n4;i++)
    sintbl[i+n2]=-sintbl[i];
}

static void make_bitrev(int n, int bitrev[])
{
  int i,j,k,n2;
  
  n2=n/2; i=0; j=0;
  for(;;){
    bitrev[i]=j;
    if(++i>=n) break;
    k=n2;
    while(k<=j){j-=k;k/=2;}
    j+=k;
  }
}

int fft(int n, vector x, vector y)
{
  static int    last_n = 0;
  static int   *bitrev = NULL;
  static SCALAR *sintbl = NULL;
  int i, j, k, ik, h, d, k2, n4, inverse;
  SCALAR t, s, c, dx, dy;

  if (n < 0) {
    n = -n;  inverse = 1;
  } else inverse = 0;
  n4 = n/4;
  if (n!=last_n || n==0) {
    last_n = n;
    if (sintbl != NULL) free(sintbl);
    if (bitrev != NULL) free(bitrev);
    if (n == 0) return 0;
    sintbl=new_vector(n+n4);
    bitrev=malloc(n*sizeof(int));
    if (sintbl == NULL || bitrev == NULL) {
      fprintf(stderr, "no memory space\n");  return 1;
    }
    make_sintbl(n, sintbl);
    make_bitrev(n, bitrev);
  }
  for (i=0;i<n;i++) {
    j=bitrev[i];
    if(i<j){
      t=x[i]; x[i]=x[j]; x[j]=t;
      t=y[i]; y[i]=y[j]; y[j]=t;
    }
  }
  for(k=1;k<n;k=k2){
    h=0; k2=k+k; d=n/k2;  
    for(j=0;j<k;j++){
      c=sintbl[h+n4];
      if(inverse)s=-sintbl[h];
      else       s= sintbl[h];
      for(i=j;i<n;i+=k2){
	ik = i+k;
	dx = s*y[ik]+c*x[ik];
	dy = c*y[ik]-s*x[ik];
	x[ik]=x[i]-dx; x[i]+=dx;
	y[ik]=y[i]-dy; y[i]+=dy;
      }
      h+=d;
    }
  }
  if(!inverse)
    for(i=0;i<n;i++){x[i]/=n;y[i]/=n;}
  return 0;
}

void hamming(int length, vector w)
{
  int i,num=length-1;
  for(i=0;i<length;++i)
    w[i]=0.54-0.46*((SCALAR)cos(2*PI*((double)i)/((double)num)));
}

void hanning(int length, vector w)
{
  int i,num=length-1;
  for(i=0;i<length;++i)
    w[i]=0.5-0.5*((SCALAR)cos(2*PI*((double)i)/((double)num)));
}

void rectangle(int length, vector w)
{
  int i;
  for(i=0;i<length;++i)w[i]=1;
}

int spl(int length, int nfft, int shift)
{ 
/*   printf("  Length of spectrogram=%d\n",1+(length-nfft)/shift); */
  return(1+(length-nfft)/shift);
}

int ispl(int spl, int nfft, int shift)
{
/*   printf("  Length of ispecgram  =%d\n",(spl-1)*shift+nfft); */
  return((spl-1)*shift+nfft);
}

void specgram(vector x, int spl, int length, int shift, int nfft, 
	      matrix spr, matrix spi)
{
  int i,j;
  vector tmpr,tmpi,w;

  printf("Making Spectrogram\n");

  w=new_vector(nfft); tmpr=new_vector(nfft); tmpi=new_vector(nfft);
  hamming(nfft, w);
  if(length<nfft){printf("data is too short\n");exit(0);}

  for(i=0;i<spl;++i){
    for(j=0;j<nfft;++j){
      tmpr[j]=x[i*shift+j]*w[j];
      tmpi[j]=0;
    }
    for(fft(nfft, tmpr, tmpi),j=0;j<nfft/2+1;++j){
      spr[j][i]=tmpr[j];
      spi[j][i]=tmpi[j];
    }
  }

  free_vector(w); free_vector(tmpr); free_vector(tmpi);
  printf("Done\n");
}
  
void ispecgram(vector x, int spl, int length, int shift, int nfft, 
	       vector w, matrix spr, matrix spi)
{
  int i,j;
  vector ova,tmpr,tmpi;

  printf("Making Inverse Spectrogram\n");

  ova=new_vector(length);tmpr=new_vector(nfft);tmpi=new_vector(nfft);

  for(i=0;i<length;++i){
    ova[i]=0; x[i]=0;
  }

  for(i=0;i<spl;++i){
    for(j=0;j<nfft/2+1;++j){
      tmpr[j]=spr[j][i];tmpi[j]=spi[j][i];
    }
    for(j=nfft/2+1;j<nfft;++j){
      tmpr[j]= tmpr[nfft-j]; tmpi[j]=-tmpi[nfft-j];
    }
    for(fft(-nfft,tmpr,tmpi),j=0;j<nfft;++j){
      ova[i*shift+j]+=w[j];
      x[i*shift+j]  +=tmpr[j];
    }
  }

  for(i=0;i<length;++i)x[i]/=ova[i];

  free_vector(ova); free_vector(tmpr); free_vector(tmpi);
  printf("Done\n");
}

int conv(int f_size, int d_size, vector fil, vector x, vector output)
{
  int t,i,d;

  d = d_size - f_size + 1;

  if(d>0)
    for(t=f_size-1;t<d_size;++t){
      for(output[t-(f_size-1)]=0,i=0;i<f_size;++i){
	output[t-(f_size-1)]+=fil[i]*x[t-i];
      }
    }
  else return(0);
  return(d);
}

