#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "wav_io.h"

int Read_matrix(int n, int m, char *fname, matrix x)
{
  FILE *fin;
  int i;

  fopen_r(fin,fname);
  for(i=0;i<n;++i)
    fread(x[i],sizeof(SCALAR),m,fin);
  return fclose(fin);
}

int Write_matrix(int n, int m, matrix x, char *fname)
{
  FILE *fout;
  int i;

  fopen_w(fout,fname);
  for(i=0;i<n;++i)
    fwrite(x[i],sizeof(SCALAR),m,fout);
  return fclose(fout);
}

unsigned long read_wav(char *fname,  wavfmt *readwav)
{
  char magic[5];
  unsigned long len=0, tmplong=0;
  short tmpshort=0;
  FILE *fp;

  magic[4]='\0';

  fopen_r(fp,fname);

  fread(magic,1,4,fp);

  if (strncmp(magic, "RIFF", 4) != 0){
    printf("Sorry, but this is not a WAVE file\n"); exit(0);
  }

  fread(&len, 4,1,fp);
  fread(magic,1,4,fp);

  if (strncmp(magic, "WAVE", 4) != 0){
    printf("Sorry, but this is not a WAVE file\n"); exit(0);
  }

  fread(magic,1,4,fp);
  if (strncmp(magic , "fmt ", 4) !=0){
    printf("I cannot fpd fmt\n"); exit(0);
  }

  fread(&len,4,1,fp);
  if ( len < 16 ){
    printf("Sorry, fmt chunk is too short\n"); exit(0);
  }

  fread(&tmpshort,2,1,fp);

  if(tmpshort==WAVE_FORMAT_PCM)
    printf("This is WAVE format.\n");
  else{
    printf("Sorry, but I cannot handle this format\n"); exit(0);
  }

  fread(&tmpshort,2,1,fp);

  if(tmpshort!=2)
    printf("Channels are %d and we cannot handle this\n",(int)tmpshort);
  printf("Channels are %d, ", (int)tmpshort);

  fread(&tmplong,4,1,fp);  printf("sampling rate is %ld Hz\n",tmplong);
  readwav->FS=tmplong;
  fread(&tmplong,4,1,fp);  printf("%ld byte/sec, ",tmplong);
  fread(&tmpshort,2,1,fp); printf("%d block align, ",tmpshort);
  fread(&tmpshort,2,1,fp); printf("%d bits/samp, \n",tmpshort);

  fread(magic,4,1,fp);
  if (strncmp(magic , "data", 4) !=0){
    printf("Sorry, missing data chunk\n"); exit(0);
  }

  fread(&len,4,1,fp);

  readwav->length=len;
  readwav->data=(short *)malloc((sizeof(short))*len/2);
  fread(readwav->data,2,len/2,fp);

  fclose(fp);
  return(len/4);
}

void write_wav(char *fname, wavfmt *writewav)
{
  unsigned long ltmp;
  unsigned short stmp;
  FILE *fp;

  fopen_w(fp,fname);

  fwrite("RIFF",1,4,fp);
  
  ltmp=writewav->length+36; fwrite(&ltmp,4,1,fp);

  fwrite("WAVE",1,4,fp); fwrite("fmt ",1,4,fp);

  ltmp=16; fwrite(&ltmp,4,1,fp);
  stmp=1;  fwrite(&stmp,2,1,fp);
  stmp=2;  fwrite(&stmp,2,1,fp);
  ltmp=writewav->FS;   fwrite(&ltmp,4,1,fp);
  ltmp=writewav->FS*4; fwrite(&ltmp,4,1,fp);
  stmp=4;  fwrite(&stmp,2,1,fp);
  stmp=16; fwrite(&stmp,2,1,fp);

  fwrite("data",1,4,fp);
  ltmp=writewav->length; fwrite(&ltmp,4,1,fp);
  fwrite(writewav->data,2,ltmp/2,fp);

  fclose(fp);
}

FILE *write_wav_head(char *fname, wavfmt *writewav)
{
  unsigned long ltmp;
  unsigned short stmp;
  FILE *fp;

  fopen_w(fp,fname);

  fwrite("RIFF",1,4,fp);
  
  ltmp=writewav->length+36; fwrite(&ltmp,4,1,fp);

  fwrite("WAVE",1,4,fp); fwrite("fmt ",1,4,fp);

  ltmp=16; fwrite(&ltmp,4,1,fp);
  stmp=1;  fwrite(&stmp,2,1,fp);
  stmp=2;  fwrite(&stmp,2,1,fp);
  ltmp=writewav->FS;   fwrite(&ltmp,4,1,fp);
  ltmp=writewav->FS*4; fwrite(&ltmp,4,1,fp);
  stmp=4;  fwrite(&stmp,2,1,fp);
  stmp=16; fwrite(&stmp,2,1,fp);

  fwrite("data",1,4,fp);
  ltmp=writewav->length; fwrite(&ltmp,4,1,fp);
/*    fwrite(writewav->data,2,ltmp/2,fp); fclose(fp); */
  return(fp);
}

int WaveData_2_DoubleMat(wavfmt fmt, matrix x)
{
  int i,m=fmt.length/4;

  new_matrix(2,m);

  for(i=0;i<m;++i){
    x[0][i]=(SCALAR)*(fmt.data+2*i)/MAX;
    x[1][i]=(SCALAR)*(fmt.data+2*i+1)/MAX;
  }
  return m;
}

int DoubleMat_2_WaveData(unsigned long n, matrix x, wavfmt *fmt)
{
  int i;

  fmt->length=2*n*(sizeof(short));
  fmt->data=(short *)malloc(fmt->length);

  for(i=0;i<n;++i){
    if(x[0][i]>=1){
      printf("data is too large and be clipped\n");
      *(fmt->data+2*i)  =MAX-1;
    }
    else if(x[0][i]<-1){
      printf("data is too large and be clipped\n");
      *(fmt->data+2*i)  =-MAX;
    }
    else *(fmt->data+2*i)  =(short)(MAX*x[0][i]);
    if(x[1][i]>=1){
      printf("data is too large and be clipped\n");
      *(fmt->data+2*i+1)  =MAX-1;
    }
    else if(x[1][i]<-1){
      printf("data is too large and be clipped\n");
      *(fmt->data+2*i+1)  =-MAX;
    }
    else *(fmt->data+2*i+1)=(short)(MAX*x[1][i]);
  }
  return(n);
}

/*
int round(SCALAR x){
  if((x-floor(x))>0.5)return(ceil(x));
  else return floor(x);
}
*/

